
#pragma once

#include <cstdio>
#include "Stack.h"


   template <class T>
   class Tree 
   {
    private:
      struct TreeNode 
      {
         T value;
         TreeNode *left, *right, *parent;
         TreeNode () { left = right = parent = nullptr; }
      };

      TreeNode *root;

      void DeleteAll(TreeNode *p); //  


    public:
      Tree () : root(nullptr) {} //  
      ~Tree() { DeleteAll(root); }

      void CreateSampleTree ();
      void Print(FILE *f = stdout) const;
      void PrintRecursive(FILE *f, const TreeNode *r) const;

//----------------------------------------------------
      class P_Iterator //    parent
      {
          Tree    *tree;                             
          TreeNode *pos;
        public:
          P_Iterator(Tree *t = 0, TreeNode *p = 0) : tree(t), pos(p) {}
          P_Iterator(const P_Iterator & i) { tree = i.tree; pos = i.pos; }
          const P_Iterator & operator=(const P_Iterator & i) { tree = i.tree; pos = i.pos; return *this; }
          bool operator!=(const P_Iterator & i) { return pos != i.pos || tree != i.tree; } 
          T & operator*() { return pos->value; } //    0! 
          const P_Iterator & operator++(); 
      };

      P_Iterator pbegin() { return P_Iterator(this, root); }
      P_Iterator pend() { return P_Iterator(this, nullptr); }

//----------------------------------------------------
      class S_Iterator //    parent,   
                       //        !!!!
      {
		  Stack<TreeNode*> st;	
          Tree    *tree;
        public:
          S_Iterator(Tree *t = 0, TreeNode *p = 0) : tree(t), st(20) { if (p) st.Push(p); }  // 20 - magic number!
          S_Iterator(const S_Iterator & i) : st(20) { tree = i.tree; st = i.st; }
          const S_Iterator & operator=(const S_Iterator & i) { tree = i.tree; st = i.st; return *this; }
          bool operator!=(const S_Iterator & i) { return st != i.st || tree != i.tree; } 
          T & operator*() { return st.Top()->value; } //    0! 
          const S_Iterator & operator++(); 
      };

      S_Iterator sbegin() { return S_Iterator(this, root); }
      S_Iterator send() { return S_Iterator(this, nullptr); }



   };


   template <>   
   void Tree<int>::PrintRecursive(FILE *f,  const typename Tree<int>::TreeNode *r) const
   {                                                 
       if (!r) return;
       fprintf(f, " %d\n", r->value);
       PrintRecursive(f, r->left);
       PrintRecursive(f, r->right);
   }

   template <>
   void Tree<int>::Print(FILE *f) const
   {                         
       fprintf(f, "Walking and printing the tree:\n");
       if (!root) { 
          fprintf(f, "empty\n");
          return; 
       }
       Tree<int>::PrintRecursive(f, root);
   }



   //     
   template <class T>
   void Tree<T>::DeleteAll(typename Tree<T>::TreeNode *p)
   {
      if (p == nullptr) return;
      DeleteAll(p->left);
      DeleteAll(p->right);
      delete p;
   }

   template <>
   void Tree<int>::DeleteAll(typename Tree<int>::TreeNode *p)
   {
      if (p == nullptr) return;
      DeleteAll(p->left);
      DeleteAll(p->right);
      delete p;
   }


   template <class T>
   const typename Tree<T>::P_Iterator & Tree<T>::P_Iterator::operator++() 
   {
      TreeNode *q;
      if (pos == 0) return *this;
      if (pos->left) { pos = pos->left; return *this; }
      if (pos->right) { pos = pos->right; return *this; }
      for (q = pos->parent; q; pos = q, q = q->parent) {
         if (q->right && q->right != pos) {
            pos = q->right;
            return *this;
         }
	  }
      pos = nullptr;
      return *this;
   }


   template <class T>
   const typename Tree<T>::S_Iterator & Tree<T>::S_Iterator::operator++() 
   {
      TreeNode *q;
      if (st.Size() == 0) return *this;
      if (st.Top()->left) { st.Push(st.Top()->left); return *this; }
      if (st.Top()->right) { st.Push(st.Top()->right); return *this; }
      for (st.Pop(q); st.Size(); st.Pop(q)) {
         if (st.Top()->right && st.Top()->right != q) {
            st.Push(st.Top()->right);
            return *this;
         }
	  }
      return *this;
   }



   //    int
   //     
   //     ..
   //
   template <>
   void Tree<int>::CreateSampleTree()
 {
      Tree<int>::TreeNode *p, *l, *r;
      root = new Tree<int>::TreeNode;
      root->value = 100;

      p = root;                                            
      l = p->left = new Tree<int>::TreeNode;
      r = p->right = new Tree<int>::TreeNode;
      l->value = 20;
      r->value = 200;
      l->parent = r->parent = p;

      p = root->left;                                            
      l = p->left = new Tree<int>::TreeNode;
      r = p->right = new Tree<int>::TreeNode;
      l->value = 10;
      r->value = 40;
      l->parent = r->parent = p;
 
      p = root->right;                                            
      l = p->left = new Tree<int>::TreeNode;
      r = p->right = new Tree<int>::TreeNode;
      l->value = 150;
      r->value = 250;
      l->parent = r->parent = p;
 
      p = root->right->left;                                            
      l = p->left = new Tree<int>::TreeNode;
      r = p->right = new Tree<int>::TreeNode;
      l->value = 120;
      r->value = 180;
      l->parent = r->parent = p;

   }



